unit zip1;

interface

uses
    classes,sysutils;

const
      magic_pk12 = $201; // : ( pk12 : pk12rec );
      magic_pk34 = $403; // : ( pk34 : pk34rec );
      magic_pk56 = $605; // : ( pk56 : pk56rec );

type
        pksign = packed record
                pk: packed array [1..2] of Char;
                pktype: WORD;
        end;
        
        pk34rec = packed record
        	version: WORD;
        	flag: WORD;
        	CompresMethod: WORD;
        	DateTime: LongWord;
        	CRC32: LongWord;
        	CompressSize: LongWord;
        	UncompressSize: LongWord;
        	FileNameLen: WORD;
        	ExtraFieldLen: WORD;
        end;

        pk12rec = packed record
        	versionmade: WORD;
        	versionneed: WORD;
        	flag: WORD;
        	CompresMethod: WORD;
        	DateTime: LongWord;
        	CRC32: LongWord;
        	CompressSize: LongWord;
        	UncompressSize: LongWord;
        	FileNameLen: WORD;
        	ExtraFieldLen: WORD;
        	FileCommentLen: WORD;
        	StartOnDisk: WORD;
        	IntFileAttrib: WORD;
        	ExtFileAttrib: LongWord;
        	locate: LongWord;
        end;

		pk56rec = packed record
			diskno: Word;
			centraldiskno: WORD;
			CentralEntries: WORD;
			TotalEntries: WORD;
			CentralSize: LongWord;
			CentralOffset: LongWord;
			ZipCommentLen: WORD;
		end;

        TPKZHeader = class
            public
                procedure ReadHeader( str : TStream ) ;
                function filenamelength : integer ;
                function headersize : integer;
                function datasize : integer;
                procedure WriteHeader( str : TStream );
                procedure modifylocation( pos : integer );
                procedure modifyfilename( str : string );
                function iscrc32: longword ;
            private
                _filename : string;
                data : packed record
                    pk : pksign;
                    case integer of
                    magic_pk12 : ( pk12 : pk12rec );
                    magic_pk34 : ( pk34 : pk34rec );
                    magic_pk56 : ( pk56 : pk56rec );
                end;
            published
                property headertype : word read data.pk.pktype;
                property filename : string read _filename write modifyfilename ;
                property pk34 : pk34rec read data.pk34;
                property pk12 : pk12rec read data.pk12;
        end;

implementation

procedure TPKZHeader.modifyfilename( str : string );
begin
    self._filename := str;
    case self.data.pk.pktype of
    magic_pk12 : self.data.pk12.FileNameLen := length( str );
    magic_pk34 : self.data.pk34.FileNameLen := length( str );
    end;
end;

procedure TPKZHeader.modifylocation( pos : integer );
begin
    if self.data.pk.pktype = magic_pk12 then self.data.pk12.locate := pos;
end;

function TPKZHeader.filenamelength : integer;
begin
    result := 0;
    case self.data.pk.pktype of
    magic_pk12 : result := self.data.pk12.FileNameLen;
    magic_pk34 : result := self.data.pk34.FileNameLen;
  //  else raise format( '%x', [ data.pk.pktype ] );
    end;
end;

function TPKZHeader.headersize : integer;
begin
    result := 0;
    case self.data.pk.pktype of
    magic_pk12 : result := sizeof(pksign) + sizeof(pk12rec) + length(_filename);
    magic_pk34 : result := sizeof(pksign) + sizeof(pk34rec) + length(_filename);
    magic_pk56 : result := sizeof(pksign) + sizeof(pk56rec);
   // else result := 0;
    end;
end;

function TPKZHeader.iscrc32 : longword;
begin
    result := 0;
    case self.data.pk.pktype of
    magic_pk12 : result := pk12.CRC32;
    magic_pk34 : result := pk34.CRC32;
   // magic_pk56 : result := pk56;
    end;
end;

procedure TPKZHeader.WriteHeader( str : TStream );
begin
    str.WriteBuffer( self.data.pk, sizeof(pksign) );
    case self.data.pk.pktype of
    magic_pk12 : begin
        str.WriteBuffer( self.data.pk12, sizeof( pk12rec) );
        str.WriteBuffer( PChar(_filename)^, length(_filename) );
        end ;
    magic_pk34 : begin
        str.WriteBuffer( self.data.pk34, sizeof( pk34rec) );
        str.WriteBuffer( PChar(_filename)^, length(_filename) );
        end;
    magic_pk56 : str.WriteBuffer( self.data.pk56, sizeof( pk56rec) );
    end;
end;

function TPKZHeader.datasize : integer;
begin
    result := 0;
    case self.data.pk.pktype of
    magic_pk12 : result := self.data.pk12.ExtraFieldLen ;
    magic_pk34 : result := self.data.pk34.ExtraFieldLen + self.data.pk34.CompressSize;
    magic_pk56 : result := self.data.pk56.ZipCommentLen ;
   // else result := 0;
    end;
end;

procedure TPKZHeader.ReadHeader( str : TStream );
    var
        buffer: array [0..256] of Char;
begin
    self.data.pk.pktype := 0;
    str.ReadBuffer( self.data.pk, sizeof(pksign) );
    case self.data.pk.pktype of
    magic_pk12 : str.ReadBuffer( self.data.pk12, sizeof(pk12rec) );
    magic_pk34 : str.ReadBuffer( self.data.pk34, sizeof(pk34rec) );
    magic_pk56 : str.ReadBuffer( self.data.pk56, sizeof(pk56rec) );
  // else data.pk.pktype := 0;
    end;
    if self.filenamelength > 0 then
    begin
        str.ReadBuffer( buffer, self.filenamelength );
        buffer[ self.filenamelength ] := #0;
        _filename := buffer;

    end

    else

    _filename := '';

end;

end.
