unit zipmodifier;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, taskdialog0, StdCtrls, ComCtrls, strutils,
  zip1,tinyhash;

type
  TZMMethod = set of (
    zmmsliceby1stfolder,
    zmmsliceby2ndfolder,
    zmmsliceby3rdfolder,
    zmmsliceby4rdfolder,
    zmmsliceby5rdfolder,
    zmmslicebyext,
    zmmslicepicknumber,
    zmmpurge1stfolder,
    zmmpurge2ndfolder,
    zmmpurge3rdfolder,
    zmmvalidatename,
    zmmextractnestzip,
    zmmpurgeduplicatebycrc
    ,zmmrenameoldzipold
    ,zmmrenamezipheading
  );

  TZipModifier = class(TTaskDialog)
  private
    _sourcefiles : TStrings;
    _destbasename : string;
  ///  purgefirstfolder : boolean;
  ///  sliceby1stfolder : boolean;
  ///  sliceby2ndfolder : boolean;
  ///  sliceby3rdfolder : boolean;
  ///  slicebyext : boolean;
  ///  purge1stfolder : boolean;
  ///  extractnextzip : boolean;
    { Private 錾 }
  protected
    function ContentRenamer( name : string ) : string;
    function ContentsPartition( name : string ) : string;
    function execute : boolean; override;
  public
    method : TZMMethod;

    procedure addsourcefile( name : string );
    constructor Create( parent : TComponent ); override;
    property sourcefile : string write addsourcefile;
    property destbasename : string write _destbasename;
    { Public 錾 }
  end;

implementation

const threadingsize : integer = 256 * 1024 ;

function TZipModifier.ContentRenamer( name : string ) : string;
    var namepart : TStrings;
begin
    namepart := TStringList.Create;
    ExtractStrings( [ '/' ], [],  PChar(name), namepart );

    if ( zmmpurge3rdfolder in method ) and ( namepart.Count >= 2 ) then
        name := ansireplacetext( name, namepart.Strings[2] + '/', '' );
    if ( zmmpurge2ndfolder in method ) and ( namepart.Count >= 3 ) then
        name := ansireplacetext( name, namepart.Strings[1] + '/', '' );
    if ( zmmpurge1stfolder in method ) and ( namepart.Count >= 2 ) then
        name := ansireplacetext( name, namepart.Strings[0] + '/', '' );

    if ( zmmvalidatename in method ) then
    begin
        name := ansireplacetext( name, ':', '' );
        name := ansireplacetext( name, ';', '' );
        name := ansireplacetext( name, '?', '' );
        name := ansireplacetext( name, '&', '' );
    end;

    result := name;
end;

function TZipModifier.ContentsPartition( name : string ) : string;
    var
        namepart, parts : TStrings;
        ii : integer;
        ext : string;

    function pickupnumber( const str : string ) : string;
      var ii : integer;
        out : string;
    begin
      out := '';
      for ii := 0 to length(str)-1 do
        if str[ii] in [ '0'..'9' ] then out := out + str[ii];
      result := out;
    end;
begin
    verbose( '>' + name );

    namepart := TStringList.Create;
    parts := TStringList.Create;
    ExtractStrings( [ '/' ], [],  PChar(name), namepart );
    if ( zmmsliceby1stfolder in method ) and ( namepart.Count >= 2 ) then
        parts.Add( namepart.Strings[0] );
    if ( zmmsliceby2ndfolder in method ) and ( namepart.Count >= 3 ) then
        parts.Add( namepart.Strings[1] );
    if ( zmmsliceby3rdfolder in method ) and ( namepart.Count >= 4 ) then
        parts.Add( namepart.Strings[2] );
    if ( zmmsliceby4rdfolder in method ) and ( namepart.Count >= 5 ) then
        parts.Add( namepart.Strings[3] );
    if ( zmmsliceby5rdfolder in method ) and ( namepart.Count >= 6 ) then
        parts.Add( namepart.Strings[4] );

    if ( zmmslicepicknumber in method ) and ( parts.Count > 0 ) then
    begin
      for ii := 0 to parts.Count-1 do
        parts.Strings[ii] := pickupnumber( parts.Strings[ii] );
    end;

    ext := ExtractFileExt( name );
    if ( zmmslicebyext in method ) and ( length(ext) > 0 ) then
        parts.add( pchar( ext )+1 );
    
    result := ansireplacestr( parts.CommaText, '"', '' );
    verbose( '<'+result );
   /// if parts.Count = 0 then result := '';
end;

constructor TZipModifier.Create( parent : TComponent );
begin
    inherited;

    _sourcefiles := TStringList.Create;
end;

procedure TZipModifier.addsourcefile( name : string );
begin
    _sourcefiles.Add( name );
end;

type
    TSlicedZipFile = class
        public
            fst: TFileStream ;
            datasize: LongWord ;
            pk12size: LongWord ;
            filecounts: WORD ;
            filename : string;
        end;

    TZipPart = class(TSlicedZipFile)
        public
            pk12fst : TFileStream;
    end;

function TZipModifier.execute : boolean;
    var
        izip : TFileStream;
        contentpos,partitions : Thash;
      ///  part0 : TSlicedizip;
        part0 : TZipPart;
        pk12fst : TFileStream;
        ii : integer;
        readsize : integer;
        pk56: pk56rec;
        filesize : integer;
        needpart : boolean;
        crcmap : thash;
 ////       pathmap : thash;
    
    procedure hugecopy( ist, ost : TStream ; writesize : integer );
        var leftbytes : integer;
    begin
        leftbytes := writesize;
        while( leftbytes > threadingsize ) and not terminate do
        begin
            yield;
            ost.CopyFrom( ist, threadingsize );
            dec( leftbytes, threadingsize );
            inc( readsize, threadingsize );
           // shortmessage( Format('%d / %d', [ readsize, filesize ] ) );
            updatepersent( readsize  / filesize );
        end;
        if ( leftbytes > 0 ) and not terminate then
        begin
            ost.CopyFrom( ist, leftbytes );
            inc( readsize, leftbytes );
        end;
    end;

    function writeopen( filename : string ) : TFileStream;
    begin
      result := TFileStream.Create( filename , fmCreate or fmShareExclusive );
    end;
    
    procedure extract( offsetpathname : string );
        var
            partname : string;
            part0 : TZipPart;
            pkz : TPKZHeader;
            ext : string;
            nestzip : boolean;
            duplicated : boolean;
        label abortexit ;
    begin
        pkz := TPKZHeader.Create;
        partname := '';

        while not terminate do
        begin
            yield;

            pkz.ReadHeader( izip );
            if pkz.headertype = magic_pk56 then goto abortexit;
            pkz.filename := offsetpathname + ContentRenamer( pkz.filename );
          //  verbose( pkz.filename );

            if needpart then partname := ContentsPartition( pkz.filename );
            ext := ExtractFileExt( pkz.filename );
            nestzip := ( zmmextractnestzip in method ) and ( ansistricomp( PCHAR(ext), '.ZIP' ) = 0 );

            if not partitions.exists(partname) then
            begin
                part0 := TZipPart.Create;
                partitions.AddObject( partname, part0 );
                if needpart
                  then part0.filename := _destbasename + ' [' + partname + '].zip'
                  else part0.filename := _destbasename + '.zip';
                part0.fst := writeopen( part0.filename + '.temp' );
                part0.pk12fst := writeopen( part0.filename + '.pk12temp' );
            end
            else part0 := ( partitions.asobject[partname] as TZipPart);


            case pkz.headertype of
                magic_pk34 :
                    begin
                        duplicated := crcmap.exists( pkz.iscrc32 );
                        if not duplicated then crcmap.addbylongword( pkz.iscrc32, nil );
                        if duplicated and ( zmmpurgeduplicatebycrc in method ) then
                        begin // skip duplicate entries if duplicated
                            verbose( '-PK34:' + pkz.filename + ' duplicated' );
                            izip.Seek( pkz.datasize, soFromCurrent )
                        end
                        else if nestzip and
                                ( pkz.pk34.UncompressSize = pkz.pk34.CompressSize ) then
                        begin // extract nested zip if nestzip
                            izip.Seek( pkz.pk34.ExtraFieldLen, soFromCurrent );
                            verbose( '+PK34:' + pkz.filename + ' extracting' );
                            extract( pkz.filename + '/' );
                        end
                        else
                        begin
                            verbose( 'PK34:' + pkz.filename + ' => ' + part0.filename );
                            if( contentpos.exists(pkz.filename)
                                and ( contentpos.asInteger[pkz.filename] <> integer(part0.datasize) ) )
                            then begin
                                izip.Destroy;
                                part0.Destroy;
                                raise Exception.Create('t@CœeقȂt@C܂') ;
                            end;
                            contentpos.AddInt( pkz.filename, part0.datasize );
                            inc( readsize , pkz.headersize );
                            inc( part0.datasize , pkz.headersize + pkz.datasize );
                            inc( part0.filecounts );
                            pkz.WriteHeader( part0.fst );
                            if pkz.datasize > 0 then begin
                                hugecopy( izip, part0.fst, pkz.datasize );
                            end;
                        end;
                    end;
                magic_pk12 :
                    begin
                        if not contentpos.exists(pkz.filename) and ( zmmpurgeduplicatebycrc in method ) then
                        begin
                            verbose( '(PK12:' + pkz.filename );
                            izip.Seek( pkz.datasize, soFromCurrent )
                        end else if nestzip and ( pkz.pk12.UncompressSize = pkz.pk12.CompressSize )  then begin
                            izip.Seek( pkz.datasize, soFromCurrent );
                        end else begin
                            verbose( 'PK12:' + pkz.filename );
                            inc( part0.pk12size , pkz.headersize + pkz.datasize );
                            inc( readsize , pkz.headersize + pkz.datasize );
                            pkz.modifylocation( contentpos.asinteger[pkz.filename] );
                            pkz.WriteHeader( part0.pk12fst );
                            if pkz.datasize > 0 then part0.pk12fst.CopyFrom( izip, pkz.datasize );
                        end;
                    end;
                end;
            try
              updatepersent( readsize / filesize );
            finally // Ȃ񂩎ɗOo񂾂Ǎ\ĂȂ̂
            end;
          //  shortmessage( Format('%d / %d -> %d %d ',[ readsize, filesize, contentpos.Count, partitions.count ] ) );

        end;
    abortexit:
        pkz.Destroy;
    end;

begin
    needpart := ( [ zmmsliceby1stfolder, zmmsliceby2ndfolder,
      zmmsliceby3rdfolder,
      zmmsliceby4rdfolder,
      zmmsliceby5rdfolder,
      zmmslicebyext ] * method ) <> [];

    contentpos := Thash.Create;
    contentpos.duplicates := dupError;
    partitions := Thash.Create;
    partitions.duplicates := dupError;
    readsize := 0;
    filesize := 0;

    crcmap := thash.Create;

    for ii := 0 to _sourcefiles.Count -1 do
    begin
        self.verbose( 'zip:' + _sourcefiles.Strings[ii] );
        izip := TFileStream.Create( _sourcefiles.Strings[ii],
            fmOpenRead or fmShareDenyWrite);
        inc( filesize, getfilesize( izip.Handle, nil ) );
        if zmmrenamezipheading in method
        then extract( extractfilename( _sourcefiles.strings[ii] ) + '/' )
        else extract( '' );
        izip.Destroy;
    end;

    if ( zmmrenameoldzipold in method ) and not terminate then
    begin
        for ii := 0 to _sourcefiles.Count -1 do
        begin
            deletefile( PChar(_sourcefiles.Strings[ii] + '.OLD' ) );
            Renamefile( _sourcefiles.Strings[ii] , _sourcefiles.Strings[ii] + '.OLD' );
            verbose( _sourcefiles.Strings[ii] + ' => .OLD' );
        end;
    end;

    for ii := 0 to partitions.Count-1 do
    begin
        part0 := partitions.Objects[ii] as TZipPart;
        if not terminate then
        begin
            part0.pk12fst.Destroy;

            pk12fst := TFileStream.Create( part0.filename + '.pk12temp' , fmOpenRead or fmShareExclusive );
            part0.fst.CopyFrom( pk12fst, 0 );
            pk12fst.Destroy;
            deletefile( PChar(part0.filename + '.pk12temp' ) );

            pk56.diskno := 0;
            pk56.centraldiskno := 0;
            pk56.TotalEntries := part0.filecounts;
            pk56.CentralEntries := part0.filecounts;
            pk56.CentralSize := part0.pk12size;
            pk56.CentralOffset := part0.datasize;
            pk56.ZipCommentLen := 0;
            part0.fst.WriteBuffer( 'PK'#5#6, 4 );
            part0.fst.WriteBuffer( pk56, sizeof(pk56) );
            part0.fst.Destroy;
            deletefile( PChar(part0.filename + '.bak' ) );
            Renamefile( part0.filename , part0.filename + '.bak' );
            Renamefile( part0.filename + '.temp', part0.filename );
        end
        else begin
            part0.fst.Destroy;
            part0.pk12fst.Destroy;
            deletefile( PChar(part0.filename + '.pk12temp' ) );
            deletefile( PChar(part0.filename + '.temp') );
        end;
        part0.Destroy;
    end;
  ///  izip.Destroy;
    if not terminate then updatepersent(100);
    verbose( Format('%d bytes %d files -> %d zips ',
                [ filesize, contentpos.Count, partitions.count ] ) );

    contentpos.Destroy;
    partitions.Destroy;

    result := not terminate;
end;

end.
 